"""App Views"""

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import render, redirect

from esi.decorators import tokens_required, token_required
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_all_characters_from_user

from ..models import General
from ..services.jobs_service import (
    get_all_jobs_for_user
)
from ..models.industry_character import IndustryCharacter
from ..helpers import fetch_tokens_for_user

from allianceauth.services.hooks import get_extension_logger
logger = get_extension_logger(__name__)


@login_required
@permission_required("milindustry.basic_access")
def dashboard(request: WSGIRequest):
    all_characters = IndustryCharacter.objects.filter(
            eve_character__in=get_all_characters_from_user(user = request.user)
    ).prefetch_related("skills", "eve_character")

    characters_slots_overview = [
        character.get_slots_overview() for character in all_characters
    ]

    aggregated_slots = {
        "manufacturing": {
            "used": sum(char.manufacturing.used for char in characters_slots_overview),
            "max": sum(char.manufacturing.max for char in characters_slots_overview),
        },
        "reaction": {
            "used": sum(char.reaction.used for char in characters_slots_overview),
            "max": sum(char.reaction.max for char in characters_slots_overview),
        },
        "research": {
            "used": sum(char.research.used for char in characters_slots_overview),
            "max": sum(char.research.max for char in characters_slots_overview),
        },
    }

    context = {
        "characters_slots_overview": characters_slots_overview,
        "aggregated_slots": aggregated_slots,
    }

    return render(request, "milindustry/dashboard.html", context)

@login_required
@permission_required("milindustry.basic_access")
def jobs_overview(request: WSGIRequest) -> HttpResponse:
    tokens = fetch_tokens_for_user(user=request.user, scopes=General.get_esi_scopes())

    industry_jobs = get_all_jobs_for_user(tokens=tokens)

    context = {
        "industry_jobs" : sorted(industry_jobs, key=lambda job: job.job_id, reverse=False),
    }
    return render(request, "milindustry/overview.html", context)


@login_required
@permission_required("milindustry.basic_access")
@token_required(scopes=General.get_esi_scopes())
def add_character(request, token) -> HttpResponse:
    logger.info(f"Token for {token.character_name} has been added")
    IndustryCharacter.register_new_character(
        new_character_id=token.character_id
    )

    return redirect("milindustry:jobs_overview")


# Django
from typing import List
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter

from ..providers import esi
from ..dataclasses.industry_job import IndustryJob

from allianceauth.services.hooks import get_extension_logger
logger = get_extension_logger(__name__)


def fetch_jobs_for_character(token: Token):
    character = EveCharacter.objects.get(character_id=token.character_id)
    logger.info(f"Fetching jobs for character {character.character_name} ({character.character_id})")

    personal_jobs = esi.client.Industry.get_characters_character_id_industry_jobs(
        character_id = character.character_id,
        token=token.valid_access_token()
    ).results()

    corporation_jobs = []
    try:
        corporation_jobs = esi.client.Industry.get_corporations_corporation_id_industry_jobs(
            corporation_id = character.corporation_id,
            token=token.valid_access_token()
        ).results()
    except Exception as e:
        if hasattr(e, "response") and getattr(e.response, "status_code", None) == 403:
            logger.info(f"The character {character.character_name} cannot read the corporation jobs - Skipping")
        else:
            raise

    own_corporation_jobs = [
        job for job in corporation_jobs
        if job['installer_id'] == character.character_id
    ]

    return personal_jobs + own_corporation_jobs


def get_all_jobs_for_user(tokens: List[Token]):
    character_token_map = {}

    for token in tokens:
        if token.character_id not in character_token_map and token.valid_access_token():
            character_token_map[token.character_id] = token

    all_jobs = []
    
    for character_id, token in character_token_map.items():
        jobs = fetch_jobs_for_character(token)
        all_jobs.extend(jobs)

    return IndustryJob.convert_from_esi_response(
        all_jobs,
        IndustryJob.resolve_type_id,
        IndustryJob.resolve_activity_id
    )



from django.db import models
from django.db.models import UniqueConstraint
from typing import Dict, List, TypedDict

from esi.models import Token
from eveuniverse.models import EveType
from allianceauth.eveonline.models import EveCharacter

from ..helpers import SkillGroups, ActivityType
from ..dataclasses.slots_overview import CharacterSlotOverview, SlotCategory
from .general import General
from ..services.jobs_service import fetch_jobs_for_character

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)

class EsiSkill(TypedDict):
    skill_id: int
    active_skill_level: int
    skillpoints_in_skill: int
    trained_skill_level: int

class EsiSkillsData(TypedDict):
    skills: List[EsiSkill]
    total_sp: int
    unallocated_sp: int

class IndustryCharacter(models.Model):
    id = models.AutoField(primary_key=True)
    eve_character = models.OneToOneField(
        EveCharacter, related_name="milindustry_industrycharacter", on_delete=models.CASCADE
    )

    def __str__(self) -> str:
        return f"{self.eve_character.character_name} (PK:{self.pk})"

    def __repr__(self) -> str:
        return f"Character(pk={self.pk}, eve_character='{self.eve_character}')"
    
    @classmethod
    def register_new_character(cls, new_character_id: int):
        from django.db import transaction

        new_character = EveCharacter.objects.get(character_id=new_character_id)

        with transaction.atomic():
            industryCharacter, _ = IndustryCharacter.objects.update_or_create(
                eve_character=new_character,
            )
        
        return
    
    def create_or_update_skills_for_character(self, esi_skills_response: EsiSkillsData) -> None:
        industry_skill_ids = (
            SkillGroups.reaction_slots
            + SkillGroups.manufacturing_slots
            + SkillGroups.research_slots
        )

        skills_to_upsert = [
            IndustryCharacterSkill(
                industry_character=self,
                skill_type_id=skill['skill_id'],
                skill_level=skill['active_skill_level']
            )
            for skill in esi_skills_response['skills']
            if skill['skill_id'] in industry_skill_ids
        ]

        for skill in skills_to_upsert:
            IndustryCharacterSkill.objects.update_or_create(
                industry_character=skill.industry_character,
                skill_type=skill.skill_type,
                defaults={'skill_level': skill.skill_level}
            )
    
    def get_slots_overview(self) -> CharacterSlotOverview:
        # Get all the IndustryCharacterSkill objects for this character, indexed by skill_type_id
        skills = {skill.skill_type_id: skill.skill_level for skill in self.skills.all()}

        def sum_skill_levels(skill_ids):
            return sum(skills.get(skill_id, 0) for skill_id in skill_ids)

        manufacturing_max = 1 + sum_skill_levels(SkillGroups.manufacturing_slots)
        reaction_max = 1 + sum_skill_levels(SkillGroups.reaction_slots)
        research_max = 1 + sum_skill_levels(SkillGroups.research_slots)

        token = (
            Token.objects.prefetch_related("scopes")
            .filter(character_id=self.eve_character.character_id)
            .require_scopes(General.get_esi_scopes())
            .require_valid()
            .first()
        )

        jobs = fetch_jobs_for_character(token=token)

        used_manufacturing = len([
            job for job in jobs if job.get("activity_id") in ActivityType.manufacturing_activities
        ])
        used_reaction = len([
            job for job in jobs if job.get("activity_id") in ActivityType.reaction_activities
        ])
        used_research = len([
            job for job in jobs if job.get("activity_id") in ActivityType.research_activities
        ])

        return CharacterSlotOverview(
            character_name=self.eve_character.character_name,
            character_id=self.eve_character.character_id,
            manufacturing = SlotCategory(
                used=used_manufacturing,
                unused=manufacturing_max - used_manufacturing,
                max=manufacturing_max,
                disabled=11 - manufacturing_max,
            ),
            reaction = SlotCategory(
                used=used_reaction,
                unused=reaction_max - used_reaction,
                max=reaction_max,
                disabled=11 - reaction_max,
            ),
            research = SlotCategory(
                used=used_research,
                unused=research_max - used_research,
                max=research_max,
                disabled=11 - research_max,
            ),
        )

class IndustryCharacterSkill(models.Model):
    industry_character = models.ForeignKey(
        IndustryCharacter,
        related_name='skills',
        on_delete=models.CASCADE
    )
    skill_type = models.ForeignKey(
        EveType,
        related_name='industry_character_skills',
        on_delete=models.CASCADE
    )
    skill_level = models.PositiveSmallIntegerField()  # Skills go from 1 to 5

    class Meta:
        constraints = [
            UniqueConstraint(fields=['industry_character', 'skill_type'], name='unique_industry_character_skill')
        ]

    def __str__(self):
        return f"{self.industry_character} - {self.skill_type} lvl {self.skill_level}"



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Example App
from milindustry import __version__


class MilindustryConfig(AppConfig):
    """App Config"""

    name = "milindustry"
    label = "milindustry"
    verbose_name = f"MIL Industry v{__version__}"

    def ready(self):
        import milindustry.signals


"""App URLs"""

# Django
from django.urls import path

# MIL Industry App
from milindustry.views import dashboard_view

app_name: str = "milindustry"

urlpatterns = [
    path("", dashboard_view.dashboard, name="dashboard"),
    path("jobs_overview", dashboard_view.jobs_overview, name="jobs_overview"),
    path("add_character", dashboard_view.add_character, name="add_character"),
]



from celery import shared_task, group

from esi.models import Token

from .models.industry_character import IndustryCharacter
from .models.general import General
from .providers import esi

from allianceauth.services.hooks import get_extension_logger
logger = get_extension_logger(__name__)

TASK_DEFAULT_KWARGS = {"time_limit": 3600, "max_retries": 3}


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def update_character_skills(self, industry_character_pk: int):
    new_character = IndustryCharacter.objects.get(pk=industry_character_pk)

    token = (
        Token.objects.prefetch_related("scopes")
        .filter(character_id=new_character.eve_character.character_id)
        .require_scopes(General.get_esi_scopes())
        .require_valid()
        .first()
    )

    if token:
        skills = esi.client.Skills.get_characters_character_id_skills(
            character_id=new_character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        logger.info(f"{len(skills['skills'])} skills retrieved for {new_character}")

        new_character.create_or_update_skills_for_character(
            esi_skills_response=skills,
        )

    else:
        logger.info(f"No valid token for {new_character}")



from typing import List

from esi.models import Token
from esi.errors import TokenError

from app_utils.allianceauth import notify_throttled

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_all_characters_from_user

from allianceauth.services.hooks import get_extension_logger
logger = get_extension_logger(__name__)

class SkillGroups:
    reaction_slots = (
        45748, #Mass Reactions
        45749, #Advanced Mass Reactions
    )

    manufacturing_slots = (
        3387, #Mass Production
        24625, #Advanced Mass Production
    )

    research_slots = (
        3406, #Laboratory Operation
        24624, #Advanced Laboratory Operation
    )

class ActivityType:
    reaction_activities = (
        9, #Reactions
        11, #Reactions2
    )

    manufacturing_activities = (
        1, #Manufacturing
    )

    research_activities = (
        3, #TE
        4, #ME
        5, #Copy
        8, #Invention
    )

def fetch_tokens_for_user(user, scopes) -> List[Token]:
    """returns valid token for a user

    Args:
    - user: User for the token retrieval
    - scopes: Provide the required scopes.

    Exceptions:
    - TokenError: If no valid token can be found
    """
    characters_for_user = get_all_characters_from_user(user=user)

    tokens: List[Token] = []
    for character in characters_for_user:
        token = (
            Token.objects.prefetch_related("scopes")
            .filter(user=user, character_id=character.character_id)
            .require_scopes(scopes)
            .require_valid()
            .first()
        )

        if token:
            tokens.append(token)
        else:
            logger.info(f"No valid token for {character.character_name}")

    return tokens



"""
App Models
Create your models in here
"""

# Django
from django.db import models


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        """Meta definitions"""

        managed = False
        default_permissions = ()
        permissions = (("basic_access", "Can access this app"),)

    @classmethod
    def get_esi_scopes(cls) -> list:
        return [
            "esi-industry.read_character_jobs.v1",
            "esi-industry.read_corporation_jobs.v1",
            "esi-characters.read_blueprints.v1",
            "esi-skills.read_skills.v1",
            "esi-corporations.read_blueprints.v1",
        ]


from dataclasses import dataclass


@dataclass
class SlotCategory:
    used: int
    unused: int
    max: int
    disabled: int


@dataclass
class CharacterSlotOverview:
    character_name: str
    character_id: int
    manufacturing: SlotCategory
    reaction: SlotCategory
    research: SlotCategory




from dataclasses import dataclass
from typing import List, Callable
from datetime import datetime

from eveuniverse.models import (
    EveIndustryActivity,
    EveType,
)

from ..providers import esi

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)


@dataclass
class IndustryJob:
    job_id: int
    activity_id: int
    activity: str
    product_type_id: int
    product_name: str
    runs: int
    start_date: datetime
    installer_id: int
    installer_name: str
    status: str
    completion_date: datetime

    def resolve_type_id(type_id: int) -> str:
        try:
            return EveType.objects.get(id=type_id).name
        except EveType.DoesNotExist:
            return f"Unknown - Type ID: {type_id}"
        
    def resolve_activity_id(type_id: int) -> str:
        try:
            return EveIndustryActivity.objects.get(id=type_id).name
        except EveType.DoesNotExist:
            return f"Unknown - Type ID: {type_id}"

    def convert_from_esi_response(
        esi_jobs: List[dict],
        resolve_type_id: Callable[[int], str],
        resolve_activity_id: Callable[[int], str]
    ):
        jobs: List[IndustryJob] = []

        installer_ids = list({job['installer_id'] for job in esi_jobs})

        if installer_ids:
            character_mapping = {
                entry["id"]: entry["name"]
                for entry in esi.client.Universe.post_universe_names(ids=installer_ids).results()
            }
        else:
            character_mapping = {}

        for job in esi_jobs:
            jobs.append(IndustryJob(
                job_id=job['job_id'],
                activity_id=job['activity_id'],
                activity=resolve_activity_id(job['activity_id']),
                product_type_id=job['product_type_id'],
                product_name=resolve_type_id(job['product_type_id']),
                runs=job['runs'],
                start_date=job['start_date'],
                installer_id=job['installer_id'],
                installer_name=character_mapping.get(job['installer_id']),
                status=job['status'],
                completion_date=job['end_date']
            ))
        
        return jobs


from django.core.management import call_command
from django.core.management.base import BaseCommand
from eveuniverse.models import EveType


class Command(BaseCommand):
    help = "Setup all needed data for MIL Industry program to operate"

    def handle(self, *args, **options):
        call_command(
            "eveuniverse_load_data",
            "types",
            "--types-enabled-sections",
            EveType.Section.INDUSTRY_ACTIVITIES,
            EveType.Section.TYPE_MATERIALS,
        )



# Generated by Django 4.2.23 on 2025-08-01 19:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0011_extend_industry_activites'),
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
        ('milindustry', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='IndustryCharacter',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('eve_character', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='milindustry_industrycharacter', to='eveonline.evecharacter')),
            ],
        ),
        migrations.CreateModel(
            name='IndustryCharacterSkill',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('skill_level', models.PositiveSmallIntegerField()),
                ('industry_character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='skills', to='milindustry.industrycharacter')),
                ('skill_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='industry_character_skills', to='eveuniverse.evetype')),
            ],
        ),
        migrations.AddConstraint(
            model_name='industrycharacterskill',
            constraint=models.UniqueConstraint(fields=('industry_character', 'skill_type'), name='unique_industry_character_skill'),
        ),
    ]



from django.dispatch import receiver
from django.db.models.signals import post_save
from .models.industry_character import IndustryCharacter

from milindustry.tasks import update_character_skills

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)

@receiver(post_save, sender=IndustryCharacter)
def on_character_created(sender, instance, created, **kwargs):
    logger.info(f"New Industry character added: {instance.eve_character.character_name} - Refreshing all data")
    if created:
        update_character_skills.delay(
            industry_character_pk=instance.pk,
        )